@@ -101,8 +101,8 @@ class AgaveStep(WorkflowStep):
             Log.an().error(msg)
             return self._fatal(msg)
 
-        # make sure app has an agave definition
-        if 'agave' not in self._app['definition']:
+        # make sure app has an agave implementation
+        if 'agave' not in self._app['implementation']:
             msg = (
                 '"agave" step class can only be instantiated with an app that'
                 ' has an "agave" definition'
@@ -131,9 +131,9 @@ class AgaveStep(WorkflowStep):
 
         """
         # make sure the source data URI has a compatible scheme (agave)
-        if self._parsed_data_uris[self._source_context]['scheme'] != 'agave':
+        if self._parsed_data_uris[self._source_context][0]['scheme'] != 'agave':
             msg = 'invalid data uri scheme for this step: {}'.format(
-                self._parsed_data_uris[self._source_context]['scheme']
+                self._parsed_data_uris[self._source_context][0]['scheme']
             )
             Log.an().error(msg)
             return self._fatal(msg)
@@ -141,28 +141,28 @@ class AgaveStep(WorkflowStep):
         # delete folder if it already exists and clean==True
         if (
                 DataManager.exists(
-                    parsed_uri=self._parsed_data_uris[self._source_context],
+                    parsed_uri=self._parsed_data_uris[self._source_context][0],
                     agave=self._agave
                 )
                 and self._clean
         ):
             if not DataManager.delete(
-                    parsed_uri=self._parsed_data_uris[self._source_context],
+                    parsed_uri=self._parsed_data_uris[self._source_context][0],
                     agave=self._agave
             ):
                 Log.a().warning(
                     'cannot delete existing data uri: %s',
-                    self._parsed_data_uris[self._source_context]['chopped_uri']
+                    self._parsed_data_uris[self._source_context][0]['chopped_uri']
                 )
 
         # create folder
         if not DataManager.mkdir(
-                parsed_uri=self._parsed_data_uris[self._source_context],
+                parsed_uri=self._parsed_data_uris[self._source_context][0],
                 recursive=True,
                 agave=self._agave
         ):
             msg = 'cannot create data uri: {}'.format(
-                self._parsed_data_uris[self._source_context]['chopped_uri']
+                self._parsed_data_uris[self._source_context][0]['chopped_uri']
             )
             Log.an().error(msg)
             return self._fatal(msg)
@@ -170,13 +170,13 @@ class AgaveStep(WorkflowStep):
         # create _log folder
         if not DataManager.mkdir(
                 uri='{}/_log'.format(
-                    self._parsed_data_uris[self._source_context]['chopped_uri']
+                    self._parsed_data_uris[self._source_context][0]['chopped_uri']
                 ),
                 recursive=True,
                 agave=self._agave
         ):
             msg = 'cannot create _log folder in data uri: {}/_log'.format(
-                self._parsed_data_uris[self._source_context]['chopped_uri']
+                self._parsed_data_uris[self._source_context][0]['chopped_uri']
             )
             Log.an().error(msg)
             return self._fatal(msg)
@@ -196,26 +196,35 @@ class AgaveStep(WorkflowStep):
             exception.
 
         """
-        # make sure map URI is compatible scheme (agave)
-        if self._parsed_map_uri['scheme'] != 'agave':
-            msg = 'invalid map uri scheme for this step: {}'.format(
-                self._parsed_map_uri['scheme']
+        combined_file_list = []
+        for uri in self._parsed_map_uris:
+            # make sure map URI is compatible scheme (agave)
+            if uri['scheme'] != 'agave':
+                msg = 'invalid map uri scheme for this step: {}'.format(
+                    uri['scheme']
+                )
+                Log.an().error(msg)
+                return self._fatal(msg)
+
+            # get file list from URI
+            file_list = DataManager.list(
+                parsed_uri=uri,
+                globstr=self._step['map']['glob'],
+                agave=self._agave
             )
-            Log.an().error(msg)
-            return self._fatal(msg)
+            if file_list is False:
+                msg = 'cannot get contents of map uri: {}'\
+                    .format(uri['chopped_uri'])
+                Log.an().error(msg)
+                return self._fatal(msg)
 
-        # list files from URI
-        file_list = DataManager.list(
-            parsed_uri=self._parsed_map_uri,
-            agave=self._agave
-        )
-        if file_list is False:
-            msg = 'cannot get contents of map uri: {}'\
-                .format(self._parsed_map_uri['chopped_uri'])
-            Log.an().error(msg)
-            return self._fatal(msg)
+            for f in file_list:
+                combined_file_list.append({
+                    'chopped_uri': uri['chopped_uri'],
+                    'filename': f
+                })
 
-        return file_list
+        return combined_file_list
 
 
     def _run_map(self, map_item):
@@ -224,7 +233,7 @@ class AgaveStep(WorkflowStep):
 
         Args:
             self: class instance.
-            map_item: map item object (item of self._map)
+            map_item: map item object (item of self._map).
 
         Returns:
             On success: True.
@@ -235,15 +244,19 @@ class AgaveStep(WorkflowStep):
         inputs = {}
         for input_key in self._app['inputs']:
             if input_key in map_item['template']:
-                inputs[input_key] = urllib.parse.quote(
-                    str(map_item['template'][input_key] or ''),
-                    safe='/:'
-                )
+                if map_item['template'][input_key]:
+                    # only include an input if the value is a non-empty string
+                    inputs[input_key] = urllib.parse.quote(
+                        str(map_item['template'][input_key]),
+                        safe='/:'
+                    )
             else:
-                inputs[input_key] = urllib.parse.quote(
-                    str(self._app['inputs'][input_key]['default'] or ''),
-                    safe='/:'
-                )
+                if self._app['inputs'][input_key]['default']:
+                    # only include an input if the value is a non-empty string
+                    inputs[input_key] = urllib.parse.quote(
+                        str(self._app['inputs'][input_key]['default']),
+                        safe='/:'
+                    )
 
         # load default app parameters, overwrite with template parameters
         parameters = {}
@@ -251,17 +264,22 @@ class AgaveStep(WorkflowStep):
             if param_key in map_item['template']:
                 parameters[param_key] = map_item['template'][param_key]
             else:
-                parameters[param_key] \
-                    = self._app['parameters'][param_key]['default']
+                if self._app['parameters'][param_key]['default'] not in [None, '']:
+                    parameters[param_key] \
+                        = self._app['parameters'][param_key]['default']
 
         # add execution method as parameter
         parameters['exec_method'] = self._step['execution']['method']
 
+        # add execution init commands if 'init' param given
+        if 'init' in self._step['execution']['parameters']:
+            parameters['exec_init'] = self._step['execution']['parameters']['init']
+
         # construct agave app template
         name = 'gf-{}-{}-{}'.format(
             str(map_item['attempt']),
-            slugify(self._step['name']),
-            slugify(map_item['template']['output'])
+            slugify(self._step['name'], regex_pattern=r'[^-a-z0-9_]+'),
+            slugify(map_item['template']['output'], regex_pattern=r'[^-a-z0-9_]+')
         )
         name = name[:62]+'..' if len(name) > 64 else name
         archive_path = '{}/{}'.format(
@@ -270,15 +288,28 @@ class AgaveStep(WorkflowStep):
         )
         app_template = {
             'name': name,
-            'appId': self._app['definition']['agave']['agave_app_id'],
+            'appId': self._app['implementation']['agave']['agave_app_id'],
             'archive': True,
             'inputs': inputs,
             'parameters': parameters,
             'archiveSystem': self._agave['parsed_archive_uri']['authority'],
             'archivePath': archive_path
         }
+        # specify processors if 'slots' param given
+        if 'slots' in self._step['execution']['parameters']:
+            app_template['processorsPerNode'] = int(
+                self._step['execution']['parameters']['slots']
+            )
+        # specify memory if 'mem' param given
+        if 'mem' in self._step['execution']['parameters']:
+            app_template['memoryPerNode'] = '{}'.format(
+                self._step['execution']['parameters']['mem']
+            )
+
         Log.some().debug(
-            "agave app template:\n%s", pprint.pformat(app_template)
+                "[step.%s]: agave app template:\n%s",
+                self._step['name'],
+                pprint.pformat(app_template)
         )
 
         # delete archive path if it exists
@@ -306,7 +337,12 @@ class AgaveStep(WorkflowStep):
             return self._fatal(msg)
 
         # log agave job id
-        Log.some().debug('agave job id: %s -> %s', map_item['template']['output'], job['id'])
+        Log.some().debug(
+            '[step.%s]: agave job id: %s -> %s',
+            self._step['name'],
+            map_item['template']['output'],
+            job['id']
+        )
 
         # record job info
         map_item['run'][map_item['attempt']]['agave_job_id'] = job['id']
@@ -421,69 +457,61 @@ class AgaveStep(WorkflowStep):
                         map_item['run'][map_item['attempt']]['hpc_job_id'] \
                             = match.group(1)
 
-                        # log hpc job id in
+                        # log hpc job id
                         Log.some().debug(
-                            'hpc job id: %s -> %s', map_item['template']['output'],
+                            '[step.%s]: hpc job id: %s -> %s',
+                            self._step['name'],
+                            map_item['template']['output'],
                             match.group(1)
                         )
 
                         break
 
+            if map_item['status'] == 'FAILED' and map_item['attempt'] < 5:
+                # retry job if not at limit
+                if not self.retry_failed(map_item):
+                    Log.a().warning(
+                        '[step.%s]: cannot retry failed agave job (%s)',
+                        self._step['name'],
+                        map_item['template']['output']
+                    )
+
         self._update_status_db(self._status, '')
 
         return True
 
 
-    def retry_failed(self):
+    def retry_failed(self, map_item):
         """
-        Check if any jobs failed or stopped.
+        Retry a job.
 
         Args:
             self: class instance.
 
         Returns:
-            If no failure: True.
-            On failure: Error message.
+            True if failed/stopped job restarted successfully
+            False if failed/stopped job not restarted due to error
 
         """
-        # check if any jobs failed or stopped
-        for map_item in self._map:
-            if (
-                    map_item['status'] == 'FAILED'
-                    or map_item['status'] == 'STOPPED'
-            ):
-                # retry the job, if not at limit
-                if map_item['attempt'] >= self._config['agave']['job_retry']:
-                    msg = (
-                        'agave job failed ({}) for step "{}", '
-                        'retries for map item "{}" reached limit of {}'
-                    ).format(
-                        map_item['run'][map_item['attempt']]['agave_job_id'],
-                        self._step['name'],
-                        map_item['filename'],
-                        self._config['agave']['job_retry']
-                    )
-                    Log.an().error(msg)
-                    return self._fatal(msg)
+        # retry job
+        Log.some().info(
+            '[step.%s]: retrying agave job (%s), attempt number %s',
+            self._step['name'],
+            map_item['template']['output'],
+            map_item['attempt']+1
+        )
 
-                # retry job
-                Log.some().info(
-                    (
-                        'agave job failed (%s) for step "%s", '
-                        'retrying map item "%s"'
-                    ),
-                    map_item['run'][map_item['attempt']]['agave_job_id'],
-                    self._step['name'],
-                    map_item['filename']
-                )
-                # add another run to list
-                map_item['attempt'] += 1
-                map_item['run'].append({})
-                if not self._run_map(map_item):
-                    msg = 'cannot re-run agave job for map item "{}"'\
-                        .format(map_item['filename'])
-                    Log.an().error(msg)
-                    return self._fatal(msg)
+        # add another run to list
+        map_item['attempt'] += 1
+        map_item['run'].append({})
+        if not self._run_map(map_item):
+            Log.a().warning(
+                '[step.%s]: cannot retry agave job (%s), attempt number %s',
+                self._step['name'],
+                map_item['template']['output'],
+                map_item['attempt']
+            )
+            return False
 
         return True
 
@@ -502,7 +530,7 @@ class AgaveStep(WorkflowStep):
         """
         # destination _log directory, common for all map items
         dest_log_dir = '{}/{}'.format(
-            self._parsed_data_uris[self._source_context]\
+            self._parsed_data_uris[self._source_context][0]\
                 ['chopped_uri'],
             '_log'
         )
@@ -512,8 +540,8 @@ class AgaveStep(WorkflowStep):
 
             # copy step output
             if not self._agave['agave_wrapper'].files_import_from_agave(
-                    self._parsed_data_uris[self._source_context]['authority'],
-                    self._parsed_data_uris[self._source_context]\
+                    self._parsed_data_uris[self._source_context][0]['authority'],
+                    self._parsed_data_uris[self._source_context][0]\
                         ['chopped_path'],
                     map_item['template']['output'],
                     '{}/{}'.format(
@@ -531,7 +559,7 @@ class AgaveStep(WorkflowStep):
                 uri=map_item['run'][map_item['attempt']]['archive_uri'],
                 agave=self._agave
             )
-            if not agave_log_list:
+            if agave_log_list is False:
                 msg = 'cannot get agave log list for step "{}"'\
                     .format(self._step['name'])
                 Log.an().error(msg)
@@ -541,10 +569,10 @@ class AgaveStep(WorkflowStep):
             for item in agave_log_list:
                 if re.match('^gf-\d*-.*\.(out|err)$', item):
                     if not self._agave['agave_wrapper'].files_import_from_agave(
-                        self._parsed_data_uris[self._source_context]\
+                        self._parsed_data_uris[self._source_context][0]\
                             ['authority'],
                         '{}/{}'.format(
-                            self._parsed_data_uris[self._source_context]\
+                            self._parsed_data_uris[self._source_context][0]\
                                 ['chopped_path'],
                             '_log'
                         ),
@@ -588,7 +616,7 @@ class AgaveStep(WorkflowStep):
                     uri=src_log_dir,
                     agave=self._agave
                 )
-                if not log_list:
+                if log_list is False:
                     msg = 'cannot get _log list for step "{}"'\
                         .format(self._step['name'])
                     Log.an().error(msg)
@@ -597,10 +625,10 @@ class AgaveStep(WorkflowStep):
                 # copy each list item
                 for item in log_list:
                     if not self._agave['agave_wrapper'].files_import_from_agave(
-                        self._parsed_data_uris[self._source_context]\
+                        self._parsed_data_uris[self._source_context][0]\
                             ['authority'],
                         '{}/{}'.format(
-                            self._parsed_data_uris[self._source_context]\
+                            self._parsed_data_uris[self._source_context][0]\
                                 ['chopped_path'],
                             '_log'
                         ),
@@ -619,21 +647,3 @@ class AgaveStep(WorkflowStep):
         self._update_status_db('FINISHED', '')
 
         return True
-
-
-    def stage(self, **kwargs):
-        """
-        Override the StageableData class method and pass Agave parameters.
-
-        Args:
-            self: class instance.
-
-        Returns:
-            On success: True.
-            On failure: False.
-
-        """
-        return super(AgaveStep, self).stage(
-            local={},
-            agave=self._agave
-        )
