@@ -2,9 +2,10 @@
 
 
 import time
+
 import requests
-import yaml
 from slugify import slugify
+import yaml
 
 from geneflow.log import Log
 from geneflow.data import DataSource, DataSourceException
@@ -12,6 +13,7 @@ from geneflow.data_manager import DataManager
 from geneflow.definition import Definition
 from geneflow.workflow_dag import WorkflowDAG, WorkflowDAGException
 from geneflow.uri_parser import URIParser
+from geneflow.extend.contexts import Contexts
 
 
 class Workflow:
@@ -43,6 +45,9 @@ class Workflow:
         self._parsed_job_work_uri = {}
         self._parsed_job_output_uri = {}
 
+        self._exec_contexts = set() # all execution contexts
+        self._data_contexts = set() # all data contexts
+
         # context-specific data and methods
         self._workflow_context = {}
 
@@ -87,6 +92,18 @@ class Workflow:
             Log.an().error(msg)
             return self._fatal(msg)
 
+        # initialize set of execution contexts
+        if not self._init_exec_context_set():
+            msg = 'cannot initialize set of execution contexts'
+            Log.an().error(msg)
+            return self._fatal(msg)
+
+        # initialize set of data contexts
+        if not self._init_data_context_set():
+            msg = 'cannot initialize set of data contexts'
+            Log.an().error(msg)
+            return self._fatal(msg)
+
         # validate all work and output URIs
         if not self._init_job_uris():
             msg = 'cannot construct and validate work and output uris'
@@ -137,14 +154,14 @@ class Workflow:
             '\n    Workflow: {}'
             '\n        Version: {}'
             '\n        Description: {}'
-            '\n        Repo: {}'
+            '\n        Git: {}'
         ).format(
             self._job['name'],
             self._job_id,
             self._workflow['name'],
             self._workflow['version'],
             self._workflow['description'],
-            self._workflow['repo_uri']
+            self._workflow['git']
         )
 
         str_rep += '\n    Inputs: '
@@ -346,7 +363,8 @@ class Workflow:
         for step_name, step in self._workflow['steps'].items():
             step['execution'] = {
                 'context': self._job['execution']['context']['default'],
-                'method': self._job['execution']['method']['default']
+                'method': self._job['execution']['method']['default'],
+                'parameters': self._job['execution']['parameters']['default']
             }
             if step_name in self._job['execution']['context']:
                 step['execution']['context'] \
@@ -354,6 +372,83 @@ class Workflow:
             if step_name in self._job['execution']['method']:
                 step['execution']['method'] \
                     = self._job['execution']['method'][step_name]
+            if step_name in self._job['execution']['parameters']:
+                step['execution']['parameters'] \
+                    = self._job['execution']['parameters'][step_name]
+
+        return True
+
+
+    def _init_exec_context_set(self):
+        """
+        Initialize set of execution contexts, which is specified by the execution.context job
+        parameters.
+
+        Args:
+            self: class instance
+
+        Returns:
+            On success: True.
+
+        """
+        # get explicit execution contexts from the job parameters
+        self._exec_contexts = set(self._job['execution']['context'].values())
+
+        # check validity of exec contexts
+        for context in self._exec_contexts:
+            if not Contexts.is_exec_context(context):
+                msg = 'invalid exec context: {}'.format(context)
+                Log.an().error(msg)
+                return self._fatal(msg)
+
+        Log.some().debug('execution contexts: %s', self._exec_contexts)
+
+        return True
+
+
+    def _init_data_context_set(self):
+        """
+        Initialize set of data contexts, which is determined by inputs and output.
+
+        Args:
+            self: class instance
+
+        Returns:
+            On success: True.
+            On failure: False.
+
+        """
+        # check input URIs for data contexts
+        for input_key in self._workflow['inputs']:
+            parsed_uri = URIParser.parse(self._workflow['inputs'][input_key]['value'][0])
+            if not parsed_uri:
+                msg = 'invalid input uri: {}'.format(
+                    self._workflow['inputs'][input_key]['value'][0]
+                )
+                Log.an().error(msg)
+                return self._fatal(msg)
+
+            self._data_contexts.add(parsed_uri['scheme'])
+
+        # add output URI data context
+        parsed_output_uri = URIParser.parse(self._job['output_uri'])
+        if not parsed_output_uri:
+            msg = 'invalid base of job output uri: {}'.format(
+                self._job['output_uri']
+            )
+            Log.an().error(msg)
+            return self._fatal(msg)
+
+        self._data_contexts.add(parsed_output_uri['scheme'])
+
+        # check validity of data contexts
+        for context in self._data_contexts:
+            if not Contexts.is_data_context(context):
+                msg = 'invalid data context: {}'.format(context)
+                Log.an().error(msg)
+                return self._fatal(msg)
+
+        Log.some().debug('data contexts: %s', self._data_contexts)
 
         return True
 
@@ -371,11 +466,21 @@ class Workflow:
 
         """
         # name of the job directory
-        job_dir = slugify(self._job['name'])
+        job_dir = slugify(self._job['name'], regex_pattern=r'[^-a-z0-9_]+')
         job_dir_hash = '{}-{}'.format(job_dir, self._job['job_id'][:8])
 
-        # validate work URI for each context
-        for context in self._job['work_uri']:
+        # validate work URI for each exec context
+        #   use the 'data_scheme' for each execution context
+        #   and place into a set to remove repeats
+        for context in {
+                Contexts.get_data_scheme_of_exec_context(con)
+                for con in self._exec_contexts
+        }:
+            # work_uri must be set for each exec_context
+            if context not in self._job['work_uri']:
+                msg = 'missing work_uri for context: {}'.format(context)
+                Log.an().error(msg)
+                return self._fatal(msg)
 
             parsed_uri = URIParser.parse(self._job['work_uri'][context])
             if not parsed_uri:
@@ -402,6 +507,7 @@ class Workflow:
 
             self._parsed_job_work_uri[context] = parsed_job_work_uri
 
+
         # validate output URI
         parsed_uri = URIParser.parse(self._job['output_uri'])
         if not parsed_uri:
@@ -411,7 +517,7 @@ class Workflow:
             Log.an().error(msg)
             return self._fatal(msg)
 
-        # append job dir to each context
+        # append job dir (hashed or not) to output uri
         full_job_output_uri = (
             '{}{}' if parsed_uri['chopped_path'] == '/' else '{}/{}'
         ).format(
@@ -446,11 +552,11 @@ class Workflow:
             On failure: False.
 
         """
-        for context in self._job['work_uri']:
-            if not self._workflow_context[context].init_data():
+        for exec_context in self._exec_contexts:
+            if not self._workflow_context[exec_context].init_data():
                 msg = (
                     'cannot initialize data for workflow context: {}'\
-                        .format(context)
+                        .format(exec_context)
                 )
                 Log.an().error(msg)
                 return self._fatal(msg)
@@ -470,7 +576,12 @@ class Workflow:
             On failure: False.
 
         """
-        for context in self._job['work_uri']:
+        # currently the union of all execution and data contexts will be used
+        # to initialize workflow contexts/classes. the reason is that all supported
+        # data contexts are also execution contexts. This may change in the future
+        # with data-only contexts (e.g., http/s). In that case, a new method
+        # (_init_data_contexts) will be added to populate a _data_context variable.
+        for context in self._exec_contexts | self._data_contexts:
 
             mod_name = '{}_workflow'.format(context)
             cls_name = '{}Workflow'.format(context.capitalize())
@@ -523,9 +634,11 @@ class Workflow:
             On failure: False.
 
         """
-        # create work URI for each context
-        for context in self._job['work_uri']:
-
+        # create work URIs. a work URI is required for each workflow context
+        for context in {
+                Contexts.mapping[exec_context]['data_scheme']
+                for exec_context in self._exec_contexts
+        }:
             if not DataManager.mkdir(
                     parsed_uri=self._parsed_job_work_uri[context],
                     recursive=True,
@@ -540,9 +653,9 @@ class Workflow:
                 Log.an().error(msg)
                 return self._fatal(msg)
 
-        # create output URI
+        # create output URI. output URI scheme must be in the set of data contexts
         output_context = self._parsed_job_output_uri['scheme']
-        if output_context not in self._workflow_context:
+        if output_context not in self._data_contexts:
             msg = 'invalid output context: {}'.format(output_context)
             Log.an().error(msg)
             return self._fatal(msg)
@@ -581,6 +694,8 @@ class Workflow:
             self._apps,
             self._parsed_job_work_uri,
             self._parsed_job_output_uri,
+            self._exec_contexts,
+            self._data_contexts,
             self._config,
             **{
                 context: self._workflow_context[context].get_context_options()\
@@ -598,6 +713,11 @@ class Workflow:
         return True
 
 
+    def _re_init(self):
+        """Reinitialize connection object."""
+        return True
+
+
     def run(self):
         """
         Run Workflow.
@@ -615,8 +735,10 @@ class Workflow:
         for node_name in self._dag.get_topological_sort():
             node = self._dag.graph().nodes[node_name]
             if node['type'] == 'input':
+
                 Log.some().debug('[%s]: staging input', node_name)
                 if not node['node'].stage(
+                        move_final=False,
                         **{
                             context: self._workflow_context[context]\
                                 .get_context_options()\
@@ -629,12 +751,18 @@ class Workflow:
 
             else: # step node
 
+                # Reinit connection to exec context
+                if not self._re_init():
+                    msg = 'cannot reinit exec context'
+                    Log.an().error(msg)
+                    return self._fatal(msg)
+
                 Log.some().info(
                     '[%s]: app: %s:%s [%s]',
                     node_name,
                     node['node']._app['name'],
                     node['node']._app['version'],
-                    node['node']._app['repo_uri']
+                    node['node']._app['git']
                 )
 
                 Log.some().debug('[%s]: iterating map uri', node_name)
@@ -657,9 +785,9 @@ class Workflow:
 
                 Log.some().debug('[%s]: all jobs complete', node_name)
 
-                # check if any jobs failed
-                if node['node'].any_failed():
-                    msg = 'jobs failed for step {}'.format(node_name)
+                # check1 if step satisfies checkpoint of all, any, or none job completion
+                if not node['node'].checkpoint():
+                    msg = 'failed checkpoint for step {}'.format(node_name)
                     Log.an().error(msg)
                     return self._fatal(msg)
 
@@ -670,7 +798,7 @@ class Workflow:
                     Log.an().error(msg)
                     return self._fatal(msg)
 
-                # stage outputs
+                # stage outputs (non-final)
                 Log.some().debug('[%s]: staging output', node_name)
                 if not node['node'].stage(
                         **{
@@ -683,8 +811,27 @@ class Workflow:
                     Log.an().error(msg)
                     return self._fatal(msg)
 
+
+        # stage final outputs
+        for node_name in self._dag.get_topological_sort():
+            node = self._dag.graph().nodes[node_name]
+            if node['type'] == 'step':
+
+                Log.some().debug('[%s]: staging final output', node_name)
+                if not node['node'].stage_final(
+                        **{
+                            context: self._workflow_context[context]\
+                                .get_context_options()\
+                            for context in self._workflow_context
+                        }
+                ):
+                    msg = 'staging final output failed for step {}'.format(node_name)
+                    Log.an().error(msg)
+                    return self._fatal(msg)
+
                 Log.some().info('[%s]: complete', node_name)
 
+
         self._update_status_db('FINISHED', '')
 
         return True
